{          MINIMUM ATTAINABLE DISTANCE FOR ABUNDANCE DATA
                           R.M.Fewster
              University of St. Andrews, Scotland, UK.

                           bamcounts.p
                           -----------
Pascal program to compute the Minimum Attainable Distance for a pair of
spatial abundance maps.  This program may be freely distributed
for non-commercial purposes.  It comes with no warranty.

Edit the program at the 3 points marked *EDIT* to give the number of
deficit sites (those with a higher count in the second map than
in the first), the number of surplus sites (those with a lower count in
the second map than in the first), and total number of sites.
These values are provided by the program countsprelim.p which must be
compiled and run first.  They are found in the file Nmismfile created by
countsprelim.p. 

After the code is edited it must be compiled before running.  Note that
the code must therefore be re-compiled every time there are new maps to
compare.  The UNIX shell-script countsh automizes this process.}

program bestattainablematch(multfile, opmismatch, output);

const Ndef=173;   {*EDIT*: number of deficit sites} 
      Nsurp=226;  {*EDIT*: number of surplus sites}
      Ntotal=400; {*EDIT*: total number of sites in each map}
      N=Ndef+Nsurp;

type
   siterange = 1..N;
   cliquesets = set of siterange;
   cliquearray = array[siterange] of cliquesets;
   headerdef = array[1..Ndef] of char;
   headersurp = array[1..Nsurp] of char;


var opmism : cliquearray;
          {opmism holds the mismatch numbers for opposite mismatches in 
          cliques for each mismatched site}
   
    cc, t_set, val_set : cliquesets;
                                     {cc is the current connected clique}
                     {t_set is the current attempt at t sites making a
                      t-siteblock (typically will contain less than t sites)}
                       {val_set is the set of pooled valent elements for
                       the sites is t_set}
  
    siteval, mult, valency, Bnumber : array[1..N] of integer;
         {siteval[i]=number of sites in opmismatch[i]}
         {mult[i]=surplus (if >0) or deficit (if <0) found in i}
         {valency[i] = total (abs) multiplicity found in opmism[i]}
         {Bnumber[i]=b if i is the b'th mismatched 0 or the b'th mismatched 1}

    clusdef : array[1..Ndef] of integer; 
    clussurp : array[1..Nsurp] of integer;
                    {clusdef[i]=hh if the site with mismatch number hh is the
                     i'th zero in the current connected clique (cluster)}

    i,j,k,t, globaldeft :integer;

    opmismatch, multfile: text;  {info output from bamprelim.p}
   
    c, h, Bnumsurp, Bnumdef : integer;

    headdef : headerdef;
    headsurp : headersurp;
    BPS, ninblock, r : integer;   {used in BAM algorithm}
    timetostop, nochains, foundsiteblock, backtosingletons : boolean;
    Nsing, Noneblk, frcount, surps, defs, singtmp, oneblktmp : integer;
                                                 {counters for complexity}
    manhattan: real;


{-----PROCEDURES------------------------------------------------------}


procedure singletons;   {to swap site-singletons}
                        {only operates on the whole matrix B and the full 
                        valency vector: swaps all sites of valency 1}
var someleft : boolean;
           l : integer;

  {----swap procedure-----}

  procedure swap(sti:integer); {to swap the singleton, sti}
  var stk, m : integer;
  begin  {swap}
    Nsing:=Nsing+1;
    singtmp:=singtmp+1;
    stk:=0;
    repeat
      stk:=stk+1;
    until stk in opmism[sti];  {found stk to swap with sti}
    {writeln('Swapping sites ',sti:2,', ',stk:2);}
    
    {swap sti and stk, adjusting valencies, sitevals, headers accordingly}
    {if valency[stk]>1 must also adjust the valencies and opmism sets of 
    the other sites valent with stk}


    if abs(mult[sti])>=abs(mult[stk]) then begin

       {all of stk's individuals are used up}
       BPS:=BPS + abs(mult[stk]);
       if siteval[stk]>1 then for m:=1 to N do
          if (m in opmism[stk]) and (m<>sti) then begin
             opmism[m]:=opmism[m]-[stk];
             valency[m]:=valency[m]-abs(mult[stk]);
             siteval[m]:=siteval[m]-1;
             if (siteval[m]=1) and (m<sti) then someleft:=true;
          end;  {if}
       valency[stk]:=-1; siteval[stk]:=-1;
       if mult[stk]>0 then begin  {stk is a surplus site}
          headsurp[Bnumber[stk]]:=' '; headdef[Bnumber[sti]]:=' ';
       end
       else begin  {stk is a deficit site}
          headdef[Bnumber[stk]]:=' '; headsurp[Bnumber[sti]]:=' ';
       end;
       mult[sti]:=mult[sti]+mult[stk]; mult[stk]:=0;

    end  {abs(mult[sti])>=abs(mult[stk])}  
    else begin   {abs(mult[sti])<abs(mult[stk])}      

       {not all of stk's individuals are used up: stk does not leave any 
       opmism sets, but some valencies might decrease by abs(mult[sti])}
       
       BPS:=BPS+abs(mult[sti]);
       if siteval[stk]>1 then for m:=1 to N do
          if (m in opmism[stk]) and (m<>sti) then
          valency[m]:=valency[m]-abs(mult[sti]);
       valency[stk]:=valency[stk]-abs(mult[sti]); 
       siteval[stk]:=siteval[stk]-1;  
       mult[stk]:=mult[stk]+mult[sti];  {multiplicities of sti and stk
                                         have opposite signs}
       if mult[sti]>0 then headsurp[Bnumber[sti]]:=' ' 
       else headdef[Bnumber[sti]]:=' ';
       mult[sti]:=0; {all of sti is used up}
       if (siteval[stk]=1) and (stk<sti) then someleft:=true;
    end;  {abs(mult[sti])<abs(mult[stk])}   

    {adjust valencies of sti}
    valency[sti]:=-1; siteval[sti]:=-1;
    opmism[stk]:=opmism[stk]-[sti];    

  end;  {swap}  

{---singletons---------}

begin  {singletons}
  someleft:=true;
  while someleft=true do begin
     l:=0; someleft:=false;
     repeat
        l:=l+1;
        if siteval[l]=1 then swap(l);
     until l=N;
  end;  {while}

 { writeln('Singletons finished: number so far = ',Nsing:2);}
end;  {singletons}

{------------------------------------}

procedure onesiteblocks;   
    {finds all sites for which abs(multiplicity)>=valency; 
     these sites are automatically minimal blocks and can be swapped 
     immediately}

var keepgoing : boolean;
           hh : integer;

  {------swap_onesiteblock procedure-------} 

  procedure swap_onesiteblock(sthh:integer);  
                        {swaps the 1-siteblock centered on site sthh}
  var v, w : integer;
  begin   {swap_onesiteblock}
 
     BPS:=BPS+valency[sthh];  {the swap available on the 1-siteblock is 
                  valency[sthh] since ALL sites valent to sthh are used up}
 
   {  writeln(' 1-siteblock: centre element ', sthh:2);}
   {  writeln('Other elements:'); }  {*}
     if mult[sthh]<0 then headdef[Bnumber[sthh]]:=' ' 
     else headsurp[Bnumber[sthh]]:=' ';
     for v:=1 to N do if v in opmism[sthh] then begin
         if mult[v]<0 then headdef[Bnumber[v]]:=' '
         else headsurp[Bnumber[v]]:=' ';
         valency[v]:=-1; siteval[v]:=-1;
         for w:=1 to N do if (w in opmism[v]) and (w<>sthh) then begin
             valency[w]:=valency[w]-abs(mult[v]); 
             siteval[w]:=siteval[w]-1;
             opmism[w]:=opmism[w]-[v];
         end; {w valent to v}
        { writeln(v); } {*}
     end;   {v}
     valency[sthh]:=-1; siteval[sthh]:=-1;
     {all sites in opmism[sthh], and sthh, are now deleted}
     
     {writeln('1-siteblock swapped: BPS=', BPS:2); }   {*}
     Noneblk:=Noneblk+1;
     oneblktmp:=oneblktmp+1;
  end;  {swap_onesiteblock sub-procedure}
 
  {------onesiteblocks------}   

begin  {onesiteblocks} 

  keepgoing:=true;
  while keepgoing=true do begin
     hh:=0; keepgoing:=false;
     repeat
        hh:=hh+1;
        if (valency[hh]>0) and (abs(mult[hh])>=valency[hh]) then begin
            swap_onesiteblock(hh);
            keepgoing:=true;
            backtosingletons:=true;  {if any 1-siteblocks are found, then have 
                                      to go back through singletons later}
        end;           {keepgoing=true unless there has been a clean 
                       sweep through all sites with no 1-siteblocks found}
     until hh=N; 
  end;   {while}
  {writeln('All 1-siteblocks done'); writeln;}  {*}
  {writeln('1-siteblocks finished: number so far = ',Noneblk:2); } {*}

end;  {onesiteblocks}   


{------------------------------------}

procedure TidyB;  

{all sites with valency 0 will be given a 
blank header, have valency and siteval changed to -1. 
Procedure then checks whether there are any more sites left to be swapped.} 
 
var l:integer;

begin  {TidyB}
   for l:=1 to N do if valency[l]=0 then begin
       valency[l]:=-1; siteval[l]:=-1;
       if mult[l]<0 then headdef[Bnumber[l]]:=' ' 
       else headsurp[Bnumber[l]]:=' ';
   end;   {l}
 
   timetostop:=true;
   l:=0;
   repeat
       l:=l+1;
       if headdef[l]='u' then timetostop:=false;
   until (headdef[l]='u') or (l=N);
               {note: if there are no 'u's among headdef[l] then there 
               can be none among headsurp[l] since these sites would then
               have valency 0 and would have been removed in previous step.}
end;   {TidyB}


{------------------------------------}

procedure find_connected_clique(site:integer);

var mark : array[1..N] of integer;
    hh : integer;
    foundclique : boolean;

begin
    for hh:=1 to N do mark[hh]:=0;
    cc:=[site];
    repeat
       foundclique:=true;
       for hh:=1 to N do if (hh in cc) and (mark[hh]=0) then begin
           mark[hh]:=1;
           if not (opmism[hh]<=cc) then begin
              cc:=cc+opmism[hh];
              foundclique:=false;  {the status of foundclique is only 
                                   changed if new elements have been 
                                                         added to cc}
           end; {if}
       end;  {hh}
    until foundclique=true;
    {the connected clique containing site is now found}

    {change the headers accordingly}
    defs:=0; surps:=0;   
    for hh:=1 to N do if hh in cc then begin
        if mult[hh]<0 then begin
             headdef[Bnumber[hh]]:='c';
             defs:=defs+1;
             clusdef[defs]:=hh;
        end 
        else begin
             headsurp[Bnumber[hh]]:='c';
             surps:=surps+1;
             clussurp[surps]:=hh;
        end;
    end;  {hh}
    if defs<Ndef then for hh:=defs+1 to Ndef do clusdef[hh]:=0;
    if surps<Nsurp then for hh:=surps+1 to Nsurp do clussurp[hh]:=0;

end;  {find_connected_clique}    


{----------------------------------}

procedure check_chains;  

{determines whether the current connected clique forms a 'chain', ie. 
whether all sites have the same valency. If so, swaps clique}

var hh, val : integer;

begin
    val:=valency[clusdef[1]];    {established the valency of a site in cc, val}

    nochains:=false; hh:=1;
    repeat
       hh:=hh+1;
       if valency[clusdef[hh]]<>val then nochains:=true;
    until (nochains=true) or (hh=defs);  {now know whether all deficits in cc 
                                                           have same valency}  
    if nochains=false then begin  {now check the surpluses in cc}
       hh:=0;
       repeat
           hh:=hh+1;
           if valency[clussurp[hh]]<>val then nochains:=true;
       until (nochains=true) or (hh=surps);  {checked all surpluses in cc}
    end;   {if}   
      {all sites in cc checked: know whether or not a chain exists (ie cc 
      is perfect) }

    if nochains=false then begin   {swap the perfect cluster cc}

       {set all headers of sites in cc to blanks, and all valencies to -1}
       {BPS must be increased by defs (=surps for perfect cluster) }

       BPS:=BPS+defs;

       for hh:=1 to defs do begin
           headdef[Bnumber[clusdef[hh]]]:=' '; 
           valency[clusdef[hh]]:=-1; siteval[clusdef[hh]]:=-1; 
           mult[clusdef[hh]]:=0;
       end;   {hh a deficit in cc}
       for hh:=1 to surps do begin
           headsurp[Bnumber[clussurp[hh]]]:=' ';
           valency[clussurp[hh]]:=-1; siteval[clussurp[hh]]:=-1;
           mult[clussurp[hh]]:=0;
       end;   {hh a surplus in cc}
{***} writeln('Perfect cluster swapped: valency=', val:3,
              ' #deficit sites=',defs:3,'  #surplus sites=',surps:3);
    end;  {if cc perfect}
end;  {check_chains} 


{---------------------------------}

procedure RestoreB;  {simply replaces any remaining 'c' headers by 'u's
                     prior to performing the singletons procedure}
var hh : integer;
begin
     for hh:=1 to defs do if headdef[Bnumber[clusdef[hh]]]='c' 
     then headdef[Bnumber[clusdef[hh]]]:='u';
     for hh:=1 to surps do if headsurp[Bnumber[clussurp[hh]]]='c'
     then headsurp[Bnumber[clussurp[hh]]]:='u';
end;  {RestoreB}



{---------------------------------------------------}


procedure find_colsiteblocks (Lheaddef:headerdef; Lheadsurp:headersurp; 
t, q, permission:integer);
                       
var nrows, r0, i, j, deft: integer;  
                                 {r0 is a row with minimum deficit in Bm}
                 Cheaddef: headerdef;
                Cheadsurp: headersurp;
 tempval, deficit, tsiteval : array[1..N] of integer;
         
       {a (t,q)-colsiteblock is a set of t rows of Bm that have summed
       multiplicities at least q greater than pooled valencies}



{-----------------------}
  procedure find_tempval;  {finds all valencies for current Bm} 
  var v,w : integer;

  begin  {find_tempval}
     nrows:=0;
{***}{writeln('tempval...'); }

     for v:=1 to N do begin
        tempval[v]:=-1;
        tsiteval[v]:=-1;
     end;
     for v:=1 to defs do begin
        tempval[clusdef[v]]:=0;
        tsiteval[clusdef[v]]:=0;
     end;
     for v:=1 to surps do begin
        tempval[clussurp[v]]:=0; 
        tsiteval[clussurp[v]]:=0;
     end;

     for v:=1 to defs do begin
       if (Lheaddef[Bnumber[clusdef[v]]]='c') then begin
          for w:=1 to surps do 
             if (Lheadsurp[Bnumber[clussurp[w]]]='c') and 
             (clussurp[w] in opmism[clusdef[v]]) then begin
                 tempval[clusdef[v]]:=tempval[clusdef[v]] 
                                       + abs(mult[clussurp[w]]);
                 tsiteval[clusdef[v]]:=tsiteval[clusdef[v]]+1;
             end;  {if w an opp. mismatch of v in cc}
       end {header 'c'}
       else begin
         tempval[clusdef[v]]:=-1;
         tsiteval[clusdef[v]]:=-1;
       end;    {header not 'c'}
     end;  {v in 1 to defs}
     
     for v:=1 to surps do begin        
       if (Lheadsurp[Bnumber[clussurp[v]]]='c') then begin
          nrows:=nrows+1;
          for w:=1 to defs do
             if (Lheaddef[Bnumber[clusdef[w]]]='c') and 
            (clusdef[w] in opmism[clussurp[v]]) then begin
                 tempval[clussurp[v]]:=tempval[clussurp[v]]
                                        + abs(mult[clusdef[w]]);
                 tsiteval[clussurp[v]]:=tsiteval[clussurp[v]]+1;
             end;   {if w an opp. mismatch of v in cc}
       end {header 'c'}
       else begin
             tempval[clussurp[v]]:=-1;  
             tsiteval[clussurp[v]]:=-1; 
       end;  {header not 'c'}
     end;  {v in 1 to surps}

     for v:=1 to N do if tempval[v]=-1 then deficit[v]:=0
     else deficit[v]:=tempval[v]-abs(mult[v]);

    {writeln('found tempval array and deficits'); 
    {tempval array and deficits found, and nrows calculated}
 
  end;   {find_tempval}

{-------------------------}
  procedure delete_rows_with_deficit_too_big;

  {if we are looking at 2-siteblocks, then we can delete any row with 
  deficit greater than the (maximum multiplicity - q) since no single site 
  will ever be able to make up this deficit}

  {this is just a short cut: ignore the fact that as one row is deleted 
  the maximum multiplicity of the remaining sites might change: that would 
  make it more trouble than it's worth}

  var v,w, maxm : integer;
 
  begin  {delete_rows_with_deficit_too_big}

  {first find maximum multiplicity of the rows}
  maxm:=0;
  for v:=1 to surps do 
    if (Lheadsurp[Bnumber[clussurp[v]]]='c') and 
    (abs(mult[clussurp[v]])>maxm) then maxm:=abs(mult[clussurp[v]]);

  {now look for rows with deficit too large}
  v:=0;
  repeat
     v:=v+1;
     if (Lheadsurp[Bnumber[clussurp[v]]]='c') and  
     (deficit[clussurp[v]]>maxm-q) then begin
        Lheadsurp[Bnumber[clussurp[v]]]:='d';
        tempval[clussurp[v]]:=-1; deficit[clussurp[v]]:=0; 
        tsiteval[clussurp[v]]:=-1;
        nrows:=nrows-1;
        for w:=1 to defs do if (Lheaddef[Bnumber[clusdef[w]]]='c') and 
        (clusdef[w] in opmism[clussurp[v]]) then begin
            tempval[clusdef[w]]:=tempval[clusdef[w]]-abs(mult[clussurp[v]]);
            tsiteval[clusdef[w]]:=tsiteval[clusdef[w]]-1;
            {don't bother updating the deficit since not needed for these 
            sites anyway}
        end;
       { writeln('Deficit too big: site ',clussurp[v]:2);  } {*}
     end;  {if}
  until (v=surps) or (nrows<t);
  {writeln('2-colsiteblock: deleted rows with deficit too big'); }  {*}
  end;  {delete_rows_with_deficit_too_big}


{-------------------------}
  procedure delete_useless_rows(nrowsreqd: integer);
  
 {if permission=1 then we delete columns that have only one entry, and 
 the associated row, IF the multiplicity of the column elt. is greater  
 than or equal to that of the row elt. Such a row can never reduce the 
 overall deficit so would not be part of a minimal colsiteblock. 
 (Permission=1 is needed otherwise we have a sub-siteblock that need not 
 be minimal.)}

 {This procedure would be more comprehensive if the sum of mult.s of 
 all those 1-columns intersecting with the row were used: the row is 
 useless if this SUM is greater than its multiplicity. But again this is 
 probably more hassle than it's worth.}

  var l : integer;
      more : boolean;  
{***} { rem, rem2 : integer; }

  procedure deleteifnecessary(col : integer);  
               {delete only if |mult(col)| >= |mult(row)| }
    var row, m : integer;
    begin               {deleteifnecessary sub-procedure}
       row:=0;
{***} { for rem:=1 to N do if rem in cc then begin 
          if mult[rem]<0 then 
          write(rem:2,' ',tsiteval[rem]:2,'  ', Lheaddef[Bnumber[rem]],'  ') 
          else 
          write(rem:2,' ',tsiteval[rem]:2,'     ',Lheadsurp[Bnumber[rem]]);  
          for rem2:=1 to N do if rem2 in opmism[rem] then write('  ',rem2:2);
          writeln;
       end; }   {***}
       repeat {find row intersecting with col}
          row:=row+1;
       until (Lheadsurp[Bnumber[clussurp[row]]]='c') and 
       (clussurp[row] in opmism[clusdef[col]]);

       if abs(mult[clussurp[row]])<=abs(mult[clusdef[col]]) then begin

          {delete the column}
         { writeln('Deleting column ', clusdef[col]:2); } {*}
          Lheaddef[Bnumber[clusdef[col]]]:='d'; 
          tempval[clusdef[col]]:=-1; tsiteval[clusdef[col]]:=-1;

          {delete the intersecting row}
          Lheadsurp[Bnumber[clussurp[row]]]:='d'; 
          tempval[clussurp[row]]:=-1; tsiteval[clussurp[row]]:=-1;
          deficit[clussurp[row]]:=0;
          nrows:=nrows-1;
         { writeln('Deleted no-hoper site ',clussurp[row]:2);  }  {*}
         { writeln('more=',more); }  {*}
 
          for m:=1 to defs do if (Lheaddef[Bnumber[clusdef[m]]]='c') and 
          (clusdef[m] in opmism[clussurp[row]]) then begin
             tempval[clusdef[m]]:=tempval[clusdef[m]]-abs(mult[clussurp[row]]);
             tsiteval[clusdef[m]]:=tsiteval[clusdef[m]]-1;
             {don't bother with the deficit for column elts}
             if (m<col) and (tsiteval[clusdef[m]]=1) then more:=true;
          end;  {m}
         { writeln('Adjusted no-hope row''s valent columns');  } {*}
       end;  {if}
      { writeln('Finished deleteifnecessary');}   {*}
    end;  {delete}

    
  begin  {delete_useless_rows}
    
    repeat
       l:=0; more:=false;
       repeat
          l:=l+1;
          if (Lheaddef[Bnumber[clusdef[l]]]='c') and 
          (tsiteval[clusdef[l]]=1) then deleteifnecessary(l);    
          if nrows<nrowsreqd then foundsiteblock:=false;
       until (l=defs) or (foundsiteblock=false);
    until (more=false) or (foundsiteblock=false);
   {writeln('deleted no-hope rows'); } {*}
  end;  {delete_useless_rows}
  

{------------------------}
  procedure delete_columns_with_no_entries; 
  var v : integer;
  begin    {delete_columns_with_no_entries}
    for v:=1 to defs do 
       if (Lheaddef[Bnumber[clusdef[v]]]='c') and 
       (tempval[clusdef[v]]=0) then begin
          Lheaddef[Bnumber[clusdef[v]]]:='d'; 
          tempval[clusdef[v]]:=-1; tsiteval[clusdef[v]]:=-1; 
       end;
    {writeln('deleted columns with no entries'); } {*}
  end;     {delete_columns_with_no_entries}


{--------------------}
  procedure find_row;
  {finds row r0 with minimum deficit in Bm, adds it to t_set and
  adds its associated columns to val_set}

   var minm, v, w : integer;
   
   begin  {find_row}
        v:=0;  
        repeat 
           v:=v+1;
        until Lheadsurp[Bnumber[clussurp[v]]]='c';   
        minm:=deficit[clussurp[v]]; r0:=clussurp[v];
                             {r0 is the MISMATCH number of the row}
        if v<surps then for w:=v+1 to surps do 
            if (Lheadsurp[Bnumber[clussurp[w]]]='c') and 
           (deficit[clussurp[w]]<minm) then begin
              r0:=clussurp[w]; minm:=deficit[clussurp[w]];
            end;  {w}
        {writeln('found row and min deficit = ', minm:2); } {*}

     {found row r0 with minimum deficit}

     Cheaddef:=Lheaddef; Cheadsurp:=Lheadsurp;
     Cheadsurp[Bnumber[r0]]:='d';   {delete r0 in new matrix C}
    
     t_set:=t_set+[r0];  {** I mean this: r0, not clussurp[r0] **}
    { writeln('Added r0 = ',r0:2,' to t_set');  } {*}
     deft:=minm; tempval[r0]:=-1; deficit[r0]:=0; tsiteval[r0]:=-1;
                                              {** I mean all this too **}

     for v:=1 to defs do if (Lheaddef[Bnumber[clusdef[v]]]='c') and 
     (clusdef[v] in opmism[r0]) then begin
         val_set:=val_set+[clusdef[v]];  {** I mean this as well **}
         Cheaddef[Bnumber[clusdef[v]]]:='d'; 
         tempval[clusdef[v]]:=-1; tsiteval[clusdef[v]]:=-1;
                                  {delete columns intersecting r0 in C}
                                   {ignore the deficit for these sites}
     end;   {v}
     {t_set now contains r0 and val_set contains its intersected columns}
    { writeln('found row: r0 = ',r0:2); }  {*}
   end;    {find_row}

  
{-----------------------------}
  
begin  {main body of find_colsiteblocks(Lheaddef, Lheadsurp, t, q, 
permission) }
  { if permission=1 then writeln('looking for ',t:2,'-colsiteblocks') 
   else writeln('looking for ',t:2,' sub-colsiteblocks'); }  {*}
   foundsiteblock:=true;
   find_tempval;
   if t=2 then delete_rows_with_deficit_too_big;
                          {now have nrows rows left}

   if t=1 then begin    {special case : t=1}
    
      {looking for one row that has mult-tempval >= q }
      i:=0; foundsiteblock:=false;
      repeat
        i:=i+1;
        if (Lheadsurp[Bnumber[clussurp[i]]]='c') and 
        (abs(mult[clussurp[i]])-tempval[clussurp[i]]>=q) then begin
           foundsiteblock:=true;
           t_set :=t_set+[clussurp[i]];
           for j:=1 to defs do if (Lheaddef[Bnumber[clusdef[j]]]='c') 
           and (clusdef[j] in opmism[clussurp[i]]) then 
                val_set:=val_set+[clusdef[j]];
        end;  {if}
        until (i=surps) or (foundsiteblock=true);
   end   {t=1}

   else begin  {general case with t>1}
      if permission=1 then delete_useless_rows(t); 
            {permission=1 iff all smaller blocks previously removed}
            {this returns a new value of foundsiteblock}
      if foundsiteblock=true then delete_columns_with_no_entries;

      if nrows<t then foundsiteblock:=false;
     { writeln('nrows=', nrows:2,' and about to start find_row...'); } {*}

      if foundsiteblock=true then begin
         find_row;  
                      {finds row r0 of Bm with minimum deficit and 
                      adds this row and its intersecting columns to 
                      t_set and val_set respectively}
                      {new matrix C now set up with r0 and its columns deleted}
                      {deficit due to inclusion of r0 is deft}

         find_colsiteblocks(Cheaddef, Cheadsurp,t-1,q+deft,0); {recursive call}

         if foundsiteblock=false then begin   {r0 is not in a (t,q)-siteblock}
             Lheadsurp[Bnumber[r0]]:='d'; tempval[r0]:=-1; deficit[r0]:=0; 
             tsiteval[r0]:=-1;          {** all these r0's are OK **}
             t_set:=t_set-[r0]; 
             for i:=1 to defs do if (Lheaddef[Bnumber[clusdef[i]]]='c') and 
             (clusdef[i] in opmism[r0]) then 
                 val_set:=val_set-[clusdef[i]];
                        {only the columns there purely by virtue of r0 
                        are deleted}
                 find_colsiteblocks(Lheaddef, Lheadsurp, t, q, permission);
         end;  {if}

      end;    {if}
  
   end;  {else}
  { if foundsiteblock=true then writeln (t:2,'-siteblock');}   {*}
end;   {find_colsiteblocks}


{------------------------------------}

procedure swap_colsiteblock;  {swap the column-siteblock found}

var v,w:integer;

begin  {swap_colsiteblock}
   for v:=1 to defs do if clusdef[v] in val_set then 
   BPS:=BPS+abs(mult[clusdef[v]]);
        {in a siteblock the swap increases by the total valency}
{  writeln('Col-siteblock found : t_set:'); }  {*}
   {swap block}
   for v:=1 to surps do if (clussurp[v] in t_set) then begin
        headsurp[Bnumber[clussurp[v]]]:=' ';
        valency[clussurp[v]]:=-1; siteval[clussurp[v]]:=-1;
        for w:=1 to defs do if (clusdef[w] in opmism[clussurp[v]]) and 
        not(clusdef[w] in val_set) then begin 
            valency[clusdef[w]]:=valency[clusdef[w]]-abs(mult[clussurp[v]]);
            siteval[clusdef[w]]:=siteval[clusdef[w]]-1;
            opmism[clusdef[w]]:=opmism[clusdef[w]]-[clussurp[v]];
        end; {w valent to v}
      { writeln(clussurp[v],'  mult=',mult[clussurp[v]]:2);}   {*}
   end;  {v a surplus site in block}

 { writeln('Val_set:'); }  {*}
   for v:=1 to defs do if (clusdef[v] in val_set) then begin
        headdef[Bnumber[clusdef[v]]]:=' ';
        valency[clusdef[v]]:=-1; siteval[clusdef[v]]:=-1;
        for w:=1 to surps do if (clussurp[w] in opmism[clusdef[v]]) and
        not(clussurp[w] in t_set) then begin
            valency[clussurp[w]]:=valency[clussurp[w]]-abs(mult[clusdef[v]]);
            siteval[clussurp[w]]:=siteval[clussurp[w]]-1;
            opmism[clussurp[w]]:=opmism[clussurp[w]]-[clusdef[v]];
        end; {w valent to v}
      { writeln(clusdef[v],'  mult=',mult[clusdef[v]]:2); } {*}
   end;  {v a deficit site in block}  

  {all sites in colsiteblock have been deleted}

  {writeln('Block swapped: BPS=',BPS:2);}  {*}
end;   {swap_colsiteblock}

{-------------------------------------------------------------------}


{---------------------------------------------------}


procedure find_rowsiteblocks (Lheaddef:headerdef; Lheadsurp:headersurp; 
t, q, permission:integer);
                       
var ncols, c0, i, j, deft: integer;  
                                 {c0 is a column with minimum deficit in Bm}
                 Cheaddef: headerdef;
                Cheadsurp: headersurp;
 tempval, deficit, tsiteval : array[1..N] of integer;
         
       {a (t,q)-rowsiteblock is a set of t columns of Bm that have summed
       multiplicities at least q greater than pooled valencies}



{-----------------------}
  procedure find_tempval;  {finds all valencies for current Bm} 
  var v,w : integer;

  begin  {find_tempval}
     ncols:=0;
{***}{writeln('tempval...'); }

     for v:=1 to N do begin
         tempval[v]:=-1;
         tsiteval[v]:=-1;
     end;
     for v:=1 to defs do begin
         tempval[clusdef[v]]:=0;
         tsiteval[clusdef[v]]:=0;
     end;
     for v:=1 to surps do begin
         tempval[clussurp[v]]:=0; 
         tsiteval[clussurp[v]]:=0;
     end;
     for v:=1 to defs do begin
       if (Lheaddef[Bnumber[clusdef[v]]]='c') then begin
          ncols:=ncols+1;
          for w:=1 to surps do 
             if (Lheadsurp[Bnumber[clussurp[w]]]='c') and 
             (clussurp[w] in opmism[clusdef[v]]) then begin
                 tempval[clusdef[v]]:=tempval[clusdef[v]] 
                                       + abs(mult[clussurp[w]]);
                 tsiteval[clusdef[v]]:=tsiteval[clusdef[v]]+1;
             end;  {if w an opp. mismatch of v in cc}
       end {header 'c'}
       else begin
             tempval[clusdef[v]]:=-1;  
             tsiteval[clusdef[v]]:=-1;
       end;  {header not 'c'}
     end;  {v in 1 to defs}
     
     for v:=1 to surps do begin        
       if (Lheadsurp[Bnumber[clussurp[v]]]='c') then begin
          for w:=1 to defs do
             if (Lheaddef[Bnumber[clusdef[w]]]='c') and 
            (clusdef[w] in opmism[clussurp[v]]) then begin
                 tempval[clussurp[v]]:=tempval[clussurp[v]]
                                        + abs(mult[clusdef[w]]);
                 tsiteval[clussurp[v]]:=tsiteval[clussurp[v]]+1;
             end;   {if w an opp. mismatch of v in cc}
       end {header 'c'}
       else begin
            tempval[clussurp[v]]:=-1; 
            tsiteval[clussurp[v]]:=-1;
       end;      {header not 'c'}
     end;  {v in 1 to surps}

     for v:=1 to N do if tempval[v]=-1 then deficit[v]:=0
     else deficit[v]:=tempval[v]-abs(mult[v]);

    {writeln('found tempval array and deficits'); 
    {tempval array and deficits found, and ncols calculated}
 
  end;   {find_tempval}

{-------------------------}
  procedure delete_cols_with_deficit_too_big;

  {if we are looking at 2-siteblocks, then we can delete any column with 
  deficit greater than the (maximum multiplicity - q) since no single site 
  will ever be able to make up this deficit}

  {this is just a short cut: ignore the fact that as one column is deleted 
  the maximum multiplicity of the remaining sites might change: that would 
  make it more trouble than it's worth}

  var v,w, maxm : integer;
 
  begin  {delete_cols_with_deficit_too_big}

  {first find maximum multiplicity of the cols}
  maxm:=0;
  for v:=1 to defs do 
    if (Lheaddef[Bnumber[clusdef[v]]]='c') and 
    (abs(mult[clusdef[v]])>maxm) then maxm:=abs(mult[clusdef[v]]);

  {now look for columns with deficit too large}
  v:=0;
  repeat
     v:=v+1;
     if (Lheaddef[Bnumber[clusdef[v]]]='c') and  
     (deficit[clusdef[v]]>maxm-q) then begin
        Lheaddef[Bnumber[clusdef[v]]]:='d';
        tempval[clusdef[v]]:=-1; deficit[clusdef[v]]:=0; 
        tsiteval[clusdef[v]]:=-1;
        ncols:=ncols-1;
        for w:=1 to surps do if (Lheadsurp[Bnumber[clussurp[w]]]='c') and 
        (clussurp[w] in opmism[clusdef[v]]) then begin
            tempval[clussurp[w]]:=tempval[clussurp[w]]-abs(mult[clusdef[v]]);
            tsiteval[clussurp[w]]:=tsiteval[clussurp[w]]-1;
            {don't bother updating the deficit since not needed for these 
            sites anyway}
        end;
        {writeln('Deficit too big: site ',clusdef[v]:2);}   {*}
     end;  {if}
  until (v=defs) or (ncols<t);
  {writeln('2-rowsiteblock: deleted columns with deficit too big');}   {*}
  end;  {delete_cols_with_deficit_too_big}


{-------------------------}
  procedure delete_useless_cols(ncolsreqd: integer);
  
 {if permission=1 then we delete rows that have only one entry, and 
 the associated column, IF the multiplicity of the row elt. is greater  
 than or equal to that of the column elt. Such a column can never reduce the 
 overall deficit so would not be part of a minimal rowsiteblock. 
 (Permission=1 is needed otherwise we have a sub-siteblock that need not 
 be minimal.)}

 {This procedure would be more comprehensive if the sum of mult.s of 
 all those 1-rows intersecting with the column were used: the column is 
 useless if this SUM is greater than its multiplicity. But again this is 
 probably more hassle than it's worth.}

  var l : integer;
      more : boolean;  

  procedure deleteifnecessary(row : integer);  
               {delete only if |mult(row)| >= |mult(col)| }
    var col, m : integer;
    begin               {deleteifnecessary sub-procedure}
       col:=0;
       repeat {find col intersecting with row}
          col:=col+1;
       until (Lheaddef[Bnumber[clusdef[col]]]='c') and 
       (clusdef[col] in opmism[clussurp[row]]);

       if abs(mult[clusdef[col]])<=abs(mult[clussurp[row]]) then begin

          {delete the row}
          {writeln('Deleting row ', clussurp[row]:2);}   {*}
          Lheadsurp[Bnumber[clussurp[row]]]:='d'; 
          tempval[clussurp[row]]:=-1; tsiteval[clussurp[row]]:=-1;

          {delete the intersecting column}
          Lheaddef[Bnumber[clusdef[col]]]:='d'; 
          tempval[clusdef[col]]:=-1; tsiteval[clusdef[col]]:=-1;
          deficit[clusdef[col]]:=0;
          ncols:=ncols-1;
         { writeln('Deleted no-hoper site ',clusdef[col]:2); }   {*}
         { writeln('more=',more); }  {*}
 
          for m:=1 to surps do if (Lheadsurp[Bnumber[clussurp[m]]]='c') and 
          (clussurp[m] in opmism[clusdef[col]]) then begin
             tempval[clussurp[m]]:=tempval[clussurp[m]]-abs(mult[clusdef[col]]);
             tsiteval[clussurp[m]]:=tsiteval[clussurp[m]]-1;
             {don't bother with the deficit for row elts}
             if (m<row) and (tsiteval[clussurp[m]]=1) then more:=true;
          end;  {m}
         { writeln('Adjusted no-hope columns''s valent rows'); }  {*}
       end;  {if}
       {writeln('Finished deleteifnecessary');}   {*}
    end;  {delete}

    
  begin  {delete_useless_cols}
    
    repeat
       l:=0; more:=false;
       repeat
          l:=l+1;
          if (Lheadsurp[Bnumber[clussurp[l]]]='c') and 
          (tsiteval[clussurp[l]]=1) then deleteifnecessary(l);    
          if ncols<ncolsreqd then foundsiteblock:=false;
       until (l=surps) or (foundsiteblock=false);
    until (more=false) or (foundsiteblock=false);
   { writeln('deleted no-hope columns');}  {*}
  end;  {delete_useless_cols}
  

{------------------------}
  procedure delete_rows_with_no_entries; 
  var v : integer;
  begin    {delete_rows_with_no_entries}
    for v:=1 to surps do 
       if (Lheadsurp[Bnumber[clussurp[v]]]='c') and 
       (tempval[clussurp[v]]=0) then begin
          Lheadsurp[Bnumber[clussurp[v]]]:='d'; 
          tempval[clussurp[v]]:=-1; tsiteval[clussurp[v]]:=-1; 
       end;
    {writeln('deleted rows with no entries');} {*}
  end;     {delete_rows_with_no_entries}


{--------------------}
  procedure find_col;
  {finds column c0 with minimum deficit in Bm, adds it to t_set and
  adds its associated rows to val_set}

   var minm, v, w : integer;
   
   begin  {find_col}
        v:=0;  
        repeat 
           v:=v+1; 
        until Lheaddef[Bnumber[clusdef[v]]]='c';   
        minm:=deficit[clusdef[v]]; c0:=clusdef[v];
                             {c0 is the MISMATCH number of the column}
        if v<defs then for w:=v+1 to defs do 
          if (Lheaddef[Bnumber[clusdef[w]]]='c') and 
          (deficit[clusdef[w]]<minm) then begin
             c0:=clusdef[w]; minm:=deficit[clusdef[w]];
          end;  {w}
       { writeln('found column and min deficit = ', minm:2); }  {*}

     {found column c0 with minimum deficit}

     Cheaddef:=Lheaddef; Cheadsurp:=Lheadsurp;
     Cheaddef[Bnumber[c0]]:='d';   {delete c0 in new matrix C}
    
     t_set:=t_set+[c0];  {** I mean this: c0, not clussurp[c0] **}
    { writeln('Added c0 = ',c0:2,' to t_set'); }  {*}
     deft:=minm; tempval[c0]:=-1; deficit[c0]:=0; tsiteval[c0]:=-1;
                                              {** I mean all this too **}

     for v:=1 to surps do if (Lheadsurp[Bnumber[clussurp[v]]]='c') and 
     (clussurp[v] in opmism[c0]) then begin
         val_set:=val_set+[clussurp[v]];  {** I mean this as well **}
         Cheadsurp[Bnumber[clussurp[v]]]:='d'; 
         tempval[clussurp[v]]:=-1; tsiteval[clussurp[v]]:=-1;
                                  {delete rows intersecting c0 in C}
                                   {ignore the deficit for these sites}
     end;   {v}
     {t_set now contains c0 and val_set contains its intersected rows}
   {  writeln('found column: c0 = ',c0:2); }  {*}
   end;    {find_col}

  
{-----------------------------}
  
begin  {main body of find_rowsiteblocks(Lheaddef, Lheadsurp, t, q, 
permission) }
  { if permission=1 then writeln('looking for ',t:2,'-rowsiteblocks') 
   else writeln('looking for ',t:2,' sub-rowsiteblocks');  } {*}
   foundsiteblock:=true;
   find_tempval;
   if t=2 then delete_cols_with_deficit_too_big;
                          {now have ncols columns left}

   if t=1 then begin    {special case : t=1}
    
      {looking for one column that has mult-tempval >= q }
      i:=0; foundsiteblock:=false;
      repeat
        i:=i+1;
        if (Lheaddef[Bnumber[clusdef[i]]]='c') and 
        (abs(mult[clusdef[i]])-tempval[clusdef[i]]>=q) then begin
           foundsiteblock:=true;
           t_set :=t_set+[clusdef[i]];
           for j:=1 to surps do if (Lheadsurp[Bnumber[clussurp[j]]]='c') 
           and (clussurp[j] in opmism[clusdef[i]]) then 
                val_set:=val_set+[clussurp[j]];
        end;  {if}
        until (i=defs) or (foundsiteblock=true);
   end   {t=1}

   else begin  {general case with t>1}
      if permission=1 then delete_useless_cols(t);   
            {permission=1 iff all smaller blocks previously removed}
            {this returns a new value of foundsiteblock}
      if foundsiteblock=true then delete_rows_with_no_entries;

      if ncols<t then foundsiteblock:=false;
     { writeln('ncols=', ncols:2,' and about to start find_col...'); } {*}

      if foundsiteblock=true then begin
         find_col;  
                      {finds column c0 of Bm with minimum deficit and 
                      adds this column and its intersecting rows to 
                      t_set and val_set respectively}
                      {new matrix C now set up with c0 and its rows deleted}
                      {deficit due to inclusion of c0 is deft}

         find_rowsiteblocks(Cheaddef, Cheadsurp,t-1,q+deft,0); {recursive call}

         if foundsiteblock=false then begin {c0 is not in a (t,q)-rowsiteblock}
             Lheaddef[Bnumber[c0]]:='d'; tempval[c0]:=-1; deficit[c0]:=0; 
             tsiteval[c0]:=-1;          {** all these c0's are OK **}
             t_set:=t_set-[c0]; 
             for i:=1 to surps do if (Lheadsurp[Bnumber[clussurp[i]]]='c') 
             and (clussurp[i] in opmism[c0]) then 
                 val_set:=val_set-[clussurp[i]];
                        {only the rows there purely by virtue of c0 
                        are deleted}
                 find_rowsiteblocks(Lheaddef, Lheadsurp, t, q, permission);
         end;  {if}

      end;    {if}
  
   end;  {else}
   {if foundsiteblock=true then writeln (t:2,'-rowsiteblock');}  {*}
end;   {find_rowsiteblocks}


{------------------------------------}

procedure swap_rowsiteblock;  {swap the row-siteblock found}

var v,w:integer;

begin  {swap_rowsiteblock}
   for v:=1 to surps do if clussurp[v] in val_set then 
   BPS:=BPS+abs(mult[clussurp[v]]);
        {in a siteblock the swap increases by the total valency}
 { writeln('Row-siteblock found : t_set:'); }  {*}
   {swap block}
   for v:=1 to defs do if (clusdef[v] in t_set) then begin
        headdef[Bnumber[clusdef[v]]]:=' ';
        valency[clusdef[v]]:=-1; siteval[clusdef[v]]:=-1;
        for w:=1 to surps do if (clussurp[w] in opmism[clusdef[v]]) and 
        not(clussurp[w] in val_set) then begin 
            valency[clussurp[w]]:=valency[clussurp[w]]-abs(mult[clusdef[v]]);
            siteval[clussurp[w]]:=siteval[clussurp[w]]-1;
            opmism[clussurp[w]]:=opmism[clussurp[w]]-[clusdef[v]];
        end; {w valent to v}
      { writeln(clusdef[v],'  mult=',mult[clusdef[v]]:2); }   {*}
   end;  {v a deficit site in block}

 { writeln('Val_set:');}  {*}
   for v:=1 to surps do if (clussurp[v] in val_set) then begin
        headsurp[Bnumber[clussurp[v]]]:=' ';
        valency[clussurp[v]]:=-1; siteval[clussurp[v]]:=-1;
        for w:=1 to defs do if (clusdef[w] in opmism[clussurp[v]]) and
        not(clusdef[w] in t_set) then begin
            valency[clusdef[w]]:=valency[clusdef[w]]-abs(mult[clussurp[v]]);
            siteval[clusdef[w]]:=siteval[clusdef[w]]-1;
            opmism[clusdef[w]]:=opmism[clusdef[w]]-[clussurp[v]];
        end; {w valent to v}
      { writeln(clussurp[v],'  mult=',mult[clussurp[v]]:2);  } {*}
   end;  {v a surplus site in block}  

  {all sites in siteblock have been deleted}

  {writeln('Block swapped: BPS=',BPS:2);}  {*}
end;   {swap_rowsiteblock}


{----------------------------------------------------------------------}


{-----THE PROGRAM------------------------------------------------------}

begin

   {read in opmism sets and multiplicities and calculate valencies}
   for i:=1 to N do opmism[i]:=[];
   reset(opmismatch); reset(multfile);

   Bnumdef:=0; Bnumsurp:=0; manhattan:=0;
   for i:=1 to N do begin
      read(multfile, mult[i]);
      manhattan:=manhattan+abs(mult[i])/Ntotal;      
      if mult[i]<0 then begin
         Bnumdef:=Bnumdef+1;
         Bnumber[i]:=Bnumdef;
      end   {i a deficit site}
      else begin
         Bnumsurp:=Bnumsurp+1;
         Bnumber[i]:=Bnumsurp;
      end;  {i a surplus site}
   end;  {i}

   for i:=1 to N do begin
      opmism[i]:=[]; valency[i]:=0;
      repeat
        read(opmismatch, j);
        if j>-1 then begin
           opmism[i]:=opmism[i]+[j];
           valency[i]:=valency[i]+abs(mult[j]);
           siteval[i]:=siteval[i]+1;
        end;
      until j=-1;
      readln(opmismatch);
   end; {i}     

{***} writeln('Data read in...');
  
{-----THE BAM ALGORITHM-------------------------------------------------}

  {---set up the headers for B---}

  {mark ALL rows and columns with a 'u' (c.f. the uneconomical B-matrix) 
  since ALL rows and columns of the economical B are used.}

  for i:=1 to Ndef do headdef[i]:='u';
  for i:=1 to Nsurp do headsurp[i]:='u';


  BPS:=0;
  Nsing:=0; frcount:=0; Noneblk:=0;
{ writeln('Ntotal  mix1  mix2  #0s   #1s   size of block   # find_row'); }

 {start the singletons <-> 1-siteblocks loop}
  repeat 
     singletons;
     TidyB;     {returns a value of timetostop}
    { writeln('So far... done singletons and BPS = ',BPS:2);}  {*}
     backtosingletons:=false;
     if timetostop=false then onesiteblocks;  
        {if any 1-siteblocks are found then backtosingletons will be true}
     {if backtosingletons=true then writeln('Going back to singletons...')
     else writeln('Finished 1-siteblocks and singletons'); }  {*}   
  until (backtosingletons=false) or (timetostop=true);

  writeln('Site-singletons: ',Nsing:2);   {*}
  writeln('1-siteblocks: ', Noneblk:2);   {*}
 { if timetostop=true then writeln('timetostop') 
   else writeln('not timetostop'); }
  
  if timetostop=false then repeat
     
     repeat    {set up connected cliques and check for chains}  
     
        {find lowest k with header 'u'}
         k:=0; timetostop:=true;
         repeat
            k:=k+1;
            if mult[k]<0 then if headdef[Bnumber[k]]='u' then 
            timetostop:=false;
            if mult[k]>0 then if headsurp[Bnumber[k]]='u' then
            timetostop:=false;
         until (timetostop=false) or (k=N);

         if timetostop=false then begin
    {*}    { writeln('finding connected clique of site ',k:2,' : elements ');}
            find_connected_clique(k);
            check_chains;  {returns a value of nochains: true or false}
           { writeln; } 
           { if nochains=false then writeln('Done a chain')
            else writeln('No more chains');  }  {*}
         end;
     until (timetostop=true) or (nochains=true);
  
    { if timetostop=false then writeln('not timetostop') 
     else writeln('timetostop');    }

     if timetostop=false then begin

       {-------the r-siteblock routine----------}
         
       r:=1;
       repeat
         r:=r+1; 
         t_set:=[]; val_set:=[]; foundsiteblock:=true;
    
         writeln('Looking for ',r:2,'-siteblocks...');  {*}
         find_colsiteblocks(headdef,headsurp,r,0,1);   

         if foundsiteblock=true then begin 
            swap_colsiteblock;
            writeln('            ',r:2,'-colsiteblock: BPS= ',BPS:2);      
         end;
         if foundsiteblock=false then begin
           { writeln('no ',r:2,'-siteblocks in columns of B');}
            t_set:=[];  val_set:=[]; foundsiteblock:=true;
            find_rowsiteblocks(headdef,headsurp,r,0,1);  
            if foundsiteblock=true then begin
                swap_rowsiteblock;
                writeln('            ',r:2,'-rowsiteblock: BPS= ', BPS:2);      
            end; 
         end;
       {  if foundsiteblock=false then 
              writeln('no ',r:2,'-siteblocks in rows of B'); }

       until foundsiteblock=true;   {must terminate eventually}
 
       {----end of the r-siteblock routine-----}
     

        RestoreB;

       {do the singletons <-> 1-siteblocks loop again}
        singtmp:=0; oneblktmp:=0;
        repeat            
            singletons;
            TidyB;     {returns a value of timetostop}
           { writeln('Done singletons again and BPS = ',BPS:2);}  {*}
            backtosingletons:=false;
            if timetostop=false then onesiteblocks;
           {if any 1-siteblocks are found then backtosingletons will be true}
           { if backtosingletons=true then 
                 writeln('Going back to singletons...')
            else writeln('Finished 1-siteblocks and singletons'); }
        until (backtosingletons=false) or (timetostop=true);    

       { if timetostop=true then writeln('timetostop') 
        else writeln('not timetostop');  }  {*}
        writeln('Site-singletons: ',singtmp:2);   {*}
        writeln('1-siteblocks: ', oneblktmp:2);   {*}

     end;    {r-siteblocks, singletons and TidyB when timetostop=false}

  until timetostop=true;      

  writeln; writeln;
  writeln('-------------------------------------------------------');
  writeln('SUMMARY:');writeln;
  writeln('Number of singletons: ',Nsing:2);
  writeln('Number of 1-siteblocks: ',Noneblk:2); writeln;
  writeln('Number of sites: ',Ntotal:2);
  writeln('Initial Manhattan distance: ', manhattan:3:2);
  writeln('Final Manhattan distance: ', manhattan-2*BPS/Ntotal:3:2);
  writeln('Best possible swap: ',BPS:2); 
  writeln; writeln;

end.
