{       MINIMUM ATTAINABLE DISTANCE ALGORITHM FOR ABUNDANCE DATA
                            R.M.Fewster
              University of St. Andrews, Scotland, UK.

                           countsprelim.p
                           --------------
Pre-processing of data for abundance maps.  This program should
be compiled and run BEFORE the main program bamcounts.p.  The program may
be freely distributed for non-commercial purposes.  It comes with no
warranty.

The clique structure is read in from the file cliquesfile.

Map data is read in from file preliminfile.  The number of surplus and
deficit sites are determined, and the associated set of swappable sites is
found for each. 

Outputs are written to the files opmismatch, multfile, Nmismfile.

Edit the code at point marked *EDIT* to set the number of sites in the
maps.} 


program bampreliminaries(cliquesfile, preliminfile, opmismatch,
multfile, Nmismfile);

const N=400;  {*EDIT*: total number of sites}

type
   siterange = 1..N;
   cliquesets = set of siterange;
   cliquearray = array[siterange] of cliquesets;
   

var clique, opmism : cliquearray;
          {clique holds the clique structure; 
           opmism holds the opposite mismatches for each mismatched site}

    mismsiteno : cliquesets;  {the site numbers of the mismatched sites}
    sitemismno : array[1..N] of integer;  {the mismatch number for each site
                                           (-1 if site is a match) }
             
    dist : array[1..2,1..N] of integer;  {maps to be compared}

    preliminfile, cliquesfile : text;
    multfile, opmismatch, Nmismfile : text;

    t, i, Nmism, Ndef, Nsurp, z, cst, j : integer;
    
    x,y,dum : integer;

{------------------------------}
function sign(a:real) :integer;
begin
    if a>0 then sign:=1 else sign:=0;
end;  {sign}


{------------------------------------------------------------}


begin

{---------- read in clique structure from cliquesfile -------------}
     reset(cliquesfile);
     for z:=1 to N do begin
       clique[z]:=[];
       repeat
         read(cliquesfile, cst);
         if cst>-1 then clique[z]:=clique[z]+[cst];
       until cst=-1;
       readln(cliquesfile);
     end; {z}


{---- find mismatches and corresponding opposite mismatches ----}
    reset(preliminfile);
    rewrite(multfile);
    rewrite(opmismatch);
    Nmism:=0; mismsiteno:=[]; Ndef:=0; Nsurp:=0;

    for t:=1 to 2 do for i:=1 to N do read(preliminfile, dist[t,i]);
    for i:=1 to N do begin 
       if dist[1,i]<>dist[2,i] then begin
         Nmism:=Nmism+1;      {so this is the Nmism'th mismatched site}
         if dist[2,i]>dist[1,i] then Nsurp:=Nsurp+1 else Ndef:=Ndef+1;
         mismsiteno:=mismsiteno+[i];  {mismsiteno contains the site 
                                     numbers of the mismatches}
         sitemismno[i]:=Nmism;   {the i'th site has mismatch number Nmism}
         writeln(multfile, dist[2,i]-dist[1,i]:3);
                 {Nmism'th row of multfile contains the multiplicity of 
                                          the Nmism'th mismatched site}
       end  {i a mismatch}
       else  {i not a mismatch}
         sitemismno[i]:=-1;
    end;  {i}       

    {so far... got the set of SITE NUMBERS of mismatched sites,     
    mismsiteno, written their multiplicities into multfile, and formed 
    the array sitemismno containing the mismatch number of each site}

    {next: need the opmismatch details}

    for i:=1 to N do if i in mismsiteno then begin
                {site number i is a mismatch}
        {find the MISMATCH NUMBERS of the opposite mismatch sites to i
        and write them to the file opmismatch:}
    
        for j:=1 to N do 
           if (j in clique[i]) and (j in mismsiteno) and 
           (sign(dist[2,j]-dist[1,j])<>sign(dist[2,i]-dist[1,i])) 
           then write(opmismatch, sitemismno[j]:7);
        write(opmismatch, '    -1'); writeln(opmismatch);

    end;  {i}
    rewrite(Nmismfile);
    writeln(Nmismfile, 'Ndef=', Ndef:3,';');
    writeln(Nmismfile, 'Nsurp=', Nsurp:3,';');
    writeln(Nmismfile, 'Ntotal=', N:3,';');
    writeln('Ndef=', Ndef:3,';');
    writeln('Nsurp=', Nsurp:3,';');
    writeln('Ntotal=', N:3,';');
    
end.

