{       MINIMUM ATTAINABLE DISTANCE ALGORITHM FOR ABUNDANCE DATA
                             R.M.Fewster
               University of St. Andrews, Scotland, UK.

                             gridcounts.p
                             ------------
Demonstration data-generation program for the Best Attainable Match
algorithm on abundance (counts) data.  The program may be freely

distributed for non-commercial purposes.  It comes with no warranty.

Generates random abundance maps on a rectangular grid with
dimensions XX by YY and orientation

              0 ___ YY
              |
              |
             XX

Clique structure is set up so that the clique of each site is its 8     
nearest neighbours, or as many of these as exist.  The clique structure is
written into the file cliquesfile.

Generates 2 abundance maps, both with counts generated according
to the integer values of a uniform distribution on the interval (lowerlim,
upperlim).  The maps are written into the file preliminfile.
Edit the program at the points marked *EDIT* to set the dimensions of the
rectangular grid, XX and YY, and the values of the interval lowerlim and
upperlim. Then compile and run.}   


program make_counts(input, preliminfile, cliquesfile);

const XX=20;         {*EDIT*: dimension 1 of rectangular grid}
      YY=20;         {*EDIT*: dimension 2 of rectangular grid}
      lowerlim = 0;  {*EDIT*: lower limit for abundance in any site}
      upperlim=200;  {*EDIT*: upper limit for abundance in any site}

var t, x, y, st, wall : integer;
    ran : real;
    preliminfile, cliquesfile : text;

begin
    wall:=seed(wallclock);
    rewrite(preliminfile);
    for t:=1 to 2 do begin
      for y:=1 to YY do begin
         for x:=1 to XX do begin
            ran:=200*random(1);
            write(preliminfile, trunc(ran):3,' ');
         end;  {x}
         writeln(preliminfile); 
      end;  {y}
      writeln(preliminfile); writeln(preliminfile);
    end;  {t}

    {---------set up cliques on grid----------}

     rewrite(cliquesfile);
     writeln(cliquesfile, '  2 ', YY+1:3,' ',YY+2:3,'  -1');
     for y:=2 to YY-1 do 
        writeln(cliquesfile,y-1:3,' ',y+1:3,' ',YY+y-1:3,' ', YY+y:3,' ',
        YY+y+1:3,'  -1');
     writeln(cliquesfile, YY-1:3,' ',2*YY-1:3,' ',2*YY:3,'  -1');
     {end of first edge}
 
     for x:=2 to XX-1 do begin
        st:=(x-1)*YY+1;  {the site number represented by (x,y)}
        writeln(cliquesfile, st-YY:3,' ',st-YY+1:3,' ',st+1:3,' ',
        st+YY:3,' ',st+YY+1:3,'  -1');  {edge site}
        for y:=2 to YY-1 do begin
           st:=(x-1)*YY+y;
           writeln(cliquesfile, st-YY-1:3,' ',st-YY:3,' ',st-YY+1:3,' ',
           st-1:3,' ',st+1:3,' ',st+YY-1:3,' ',st+YY:3,' ',st+YY+1:3,'   -1');
        end; {y}
        st:=x*YY;
        writeln(cliquesfile, st-YY-1:3,' ',st-YY:3,' ',st-1:3,' ',
        st+YY-1:3,' ',st+YY:3,'  -1');  {edge site}
     end; {x}

     st:=(XX-1)*YY+1;
     writeln(cliquesfile,  st-YY:3,' ',st-YY+1:3,' ',st+1:3,'  -1');
     for y:=2 to YY-1 do begin
        st:=(XX-1)*YY+y;
        writeln(cliquesfile, st-YY-1:3,' ',st-YY:3,' ',st-YY+1:3,' ',
        st-1:3,' ',st+1:3,'   -1');
     end;  {y}
     writeln(cliquesfile, (XX-1)*YY-1:3,' ',(XX-1)*YY:3,' ',XX*YY-1:3,'  -1');
     {end of last edge}     

end.
